"""Base RPC Provider."""
from typing import Any

from ..types import RPCMethod, RPCResponse


class BaseProvider:
    """Base class for RPC providers to implement."""

    def make_request(self, method: RPCMethod, *params: Any) -> RPCResponse:
        """Make a request ot the rpc endpoint."""
        raise NotImplementedError("Providers must implement this method")

    def is_connected(self) -> bool:
        """Health check."""
        raise NotImplementedError("Providers must implement this method")



"""Solana.py is the Python API for interfacing with the `Solana JSON RPC <https://docs.solana.com/apps/jsonrpc-api/>`_.

Installation
------------

.. highlight:: sh
.. code-block:: sh

    pip install solana

Usage
-----

General usage:

.. highlight:: py
.. code-block:: py

    import solana

API Client:

.. highlight:: py
.. code-block:: py

    from solana.rpc.api import Client

    http_client = Client("https://devnet.solana.com")

"""



"""RPC types."""
from typing import Any, NewType, Union

from typing_extensions import Literal, TypedDict  # noqa: F401

URI = NewType("URI", str)
RPCMethod = NewType("RPCMethod", str)


class RPCError(TypedDict):
    """RPC error."""

    code: int
    """HTTP status code."""
    message: str
    """Error message."""


class RPCResponse(TypedDict, total=False):
    """RPC Response."""

    error: Union[RPCError, str]
    """RPC error."""
    id: int
    """Request ID."""
    jsonrpc: Literal["2.0"]
    """Protocol."""
    result: Any
    """Response results."""



"""Type for blockhash.

>>> # An arbitrary base58 encoded blockhash:
>>> Blockhash("EETubP5AKHgjPAhzPAFcb8BAY1hMH639CWCFTqi3hq1k")
'EETubP5AKHgjPAhzPAFcb8BAY1hMH639CWCFTqi3hq1k'
"""
from typing import NewType

Blockhash = NewType("Blockhash", str)



"""Library to interface with system programs."""
from __future__ import annotations

from typing import Any, List, NamedTuple, Union

from solana.instruction import InstructionLayout, decode_data, encode_data
from solana.publickey import PublicKey
from solana.transaction import AccountMeta, Transaction, TransactionInstruction

# Instruction Indices
_CREATE = 0
_ASSIGN = 1
_TRANSFER = 2
_CREATE_WITH_SEED = 3
_ADVANCE_NONCE_ACCOUNT = 4
_WITHDRAW_NONCE_ACCOUNT = 5
_INITIALZE_NONCE_ACCOUNT = 6
_AUTHORIZE_NONCE_ACCOUNT = 7
_ALLOCATE = 8
_ALLOCATE_WITH_SEED = 9
_ASSIGN_WITH_SEED = 10


# Instruction Params
class CreateAccountParams(NamedTuple):
    """Create account system transaction params."""

    from_pubkey: PublicKey
    """"""
    new_account_pubkey: PublicKey
    """"""
    lamports: int
    """"""
    space: int
    """"""
    program_id: PublicKey
    """"""


class TransferParams(NamedTuple):
    """Transfer system transaction params."""

    from_pubkey: PublicKey
    """"""
    to_pubkey: PublicKey
    """"""
    lamports: int
    """"""


class AssignParams(NamedTuple):
    """Assign system transaction params."""

    account_pubkey: PublicKey
    """"""
    program_id: PublicKey
    """"""


class CreateAccountWithSeedParams(NamedTuple):
    """Create account with seed system transaction params."""

    from_pubkey: PublicKey
    """"""
    new_account_pubkey: PublicKey
    """"""
    base_pubkey: PublicKey
    """"""
    seed: str
    """"""
    lamports: int
    """"""
    space: int
    """"""
    program_id: PublicKey
    """"""


class CreateNonceAccountParams(NamedTuple):
    """Create nonce account system transaction params.

    :param from_pubkey: PublicKey

    :param nonce_pubkey: PublicKey

    :param authorized_pubkey: PublicKey

    :param lamports: int
    """

    from_pubkey: PublicKey
    nonce_pubkey: PublicKey
    authorized_pubkey: PublicKey
    lamports: int


class CreateNonceAccountWithSeedParams(NamedTuple):
    """Create nonce account with seed system transaction params."""

    from_pubkey: PublicKey
    """"""
    nonce_pubkey: PublicKey
    """"""
    authorized_pubkey: PublicKey
    """"""
    lamports: int
    """"""
    base_pubkey: PublicKey
    """"""
    seed: str
    """"""


class InitializeNonceParams(NamedTuple):
    """Initialize nonce account system instruction params."""

    nonce_pubkey: PublicKey
    """"""
    authorized_pubkey: PublicKey
    """"""


class AdvanceNonceParams(NamedTuple):
    """Advance nonce account system instruction params."""

    nonce_pubkey: PublicKey
    """"""
    authorized_pubkey: PublicKey
    """"""


class WithdrawNonceParams(NamedTuple):
    """Withdraw nonce account system transaction params."""

    nonce_pubkey: PublicKey
    """"""
    authorized_pubkey: PublicKey
    """"""
    to_pubkey: PublicKey
    """"""
    lamports: int
    """"""


class AuthorizeNonceParams(NamedTuple):
    """Authorize nonce account system transaction params.

    :param nonce_pubkey: PublicKey

    :param authorized_pubkey: PublicKey

    :param new_authorized_pubkey: PublicKey
    """

    nonce_pubkey: PublicKey
    """"""
    authorized_pubkey: PublicKey
    """"""
    new_authorized_pubkey: PublicKey
    """"""


class AllocateParams(NamedTuple):
    """Allocate account with seed system transaction params."""

    account_pubkey: PublicKey
    """"""
    space: int
    """"""


class AllocateWithSeedParams(NamedTuple):
    """Allocate account with seed system transaction params."""

    account_pubkey: PublicKey
    """"""
    base_pubkey: PublicKey
    """"""
    seed: str
    """"""
    space: int
    """"""
    program_id: PublicKey
    """"""


class AssignWithSeedParams(NamedTuple):
    """Assign account with seed system transaction params."""

    account_pubkey: PublicKey
    """"""
    base_pubkey: PublicKey
    """"""
    seed: str
    """"""
    program_id: PublicKey
    """"""


SYSTEM_INSTRUCTION_LAYOUTS: List[InstructionLayout] = [
    InstructionLayout(idx=_CREATE, fmt="<Iqq32s"),
    InstructionLayout(idx=_ASSIGN, fmt="<I32s"),
    InstructionLayout(idx=_TRANSFER, fmt="<Iq"),
    InstructionLayout(idx=_CREATE_WITH_SEED, fmt=""),
    InstructionLayout(idx=_ADVANCE_NONCE_ACCOUNT, fmt="<I"),
    InstructionLayout(idx=_WITHDRAW_NONCE_ACCOUNT, fmt="<Iq"),
    InstructionLayout(idx=_INITIALZE_NONCE_ACCOUNT, fmt="<I32s"),
    InstructionLayout(idx=_AUTHORIZE_NONCE_ACCOUNT, fmt="<I32s"),
    InstructionLayout(idx=_ALLOCATE, fmt="<I32s"),
    InstructionLayout(idx=_ALLOCATE_WITH_SEED, fmt=""),
    InstructionLayout(idx=_ASSIGN_WITH_SEED, fmt=""),
]


def __check_key_length(keys: List[Any], expected_length: int) -> None:
    if len(keys) < expected_length:
        raise ValueError(f"invalid instruction: found {len(keys)} keys, expected at least {expected_length}")


def __check_program_id(program_id: PublicKey) -> None:
    if program_id != sys_program_id():
        raise ValueError("invalid instruction: programId is not SystemProgram")


def decode_instruction_layout(instruction: TransactionInstruction) -> InstructionLayout:
    """Decode a system instruction and retrieve the instruction layout."""
    raise NotImplementedError("decode_instruction_layout not implemented")


def decode_create_account(instruction: TransactionInstruction) -> CreateAccountParams:
    """Decode a create account system instruction and retrieve the instruction params."""
    raise NotImplementedError("decode_create_account not implemented")


def decode_transfer(instruction: TransactionInstruction) -> TransferParams:
    """Decode a transfer system instruction and retrieve the instruction params.

    >>> from solana.publickey import PublicKey
    >>> sender, reciever = PublicKey(1), PublicKey(2)
    >>> transfer_tx = transfer(
    ...     TransferParams(from_pubkey=sender, to_pubkey=reciever, lamports=1000)
    ... )
    >>> decode_transfer(transfer_tx.instructions[0])
    TransferParams(from_pubkey=11111111111111111111111111111112, to_pubkey=11111111111111111111111111111113, lamports=1000)
    """  # pylint: disable=line-too-long # noqa: E501
    __check_program_id(instruction.program_id)
    __check_key_length(instruction.keys, 2)

    layout = SYSTEM_INSTRUCTION_LAYOUTS[_TRANSFER]
    data = decode_data(layout, instruction.data)

    return TransferParams(
        from_pubkey=instruction.keys[0].pubkey, to_pubkey=instruction.keys[1].pubkey, lamports=data[1]
    )


def decode_allocate(instruction: TransactionInstruction) -> AllocateParams:
    """Decode an allocate system instruction and retrieve the instruction params."""
    raise NotImplementedError("decode_allocate not implemented")


def decode_allocate_with_seed(instruction: TransactionInstruction) -> AllocateWithSeedParams:
    """Decode an allocate with seed system instruction and retrieve the instruction params."""
    raise NotImplementedError("decode_allocate_with_seed not implemented")


def decode_assign(instruction: TransactionInstruction) -> AssignParams:
    """Decode an assign system instruction and retrieve the instruction params."""
    raise NotImplementedError("decode_assign not implemented")


def decode_assign_with_seed(instruction: TransactionInstruction) -> AssignWithSeedParams:
    """Decode an assign system with seed instruction and retrieve the instruction params."""
    raise NotImplementedError("decode_assign_with_seed not implemented")


def decode_create_with_seed(instruction: TransactionInstruction) -> CreateAccountWithSeedParams:
    """Decode a create account with seed system instruction and retrieve the instruction params."""
    raise NotImplementedError("decode_create_with_seed not implemented")


def decode_nonce_initialize(instruction: TransactionInstruction) -> InitializeNonceParams:
    """Decode a nonce initialize system instruction and retrieve the instruction params."""
    raise NotImplementedError("decode_nonce_initialize not implemented")


def decode_nonce_advance(instruction: TransactionInstruction) -> AdvanceNonceParams:
    """Decode a nonce advance system instruction and retrieve the instruction params."""
    raise NotImplementedError("decode_nonce_advance not implemented")


def decode_nonce_withdraw(instruction: TransactionInstruction) -> WithdrawNonceParams:
    """Decode a nonce withdraw system instruction and retrieve the instruction params."""
    raise NotImplementedError("decode_nonce_withdraw not implemented")


def decode_nonce_authorize(instruction: TransactionInstruction) -> AuthorizeNonceParams:
    """Decode a nonce authorize system instruction and retrieve the instruction params."""
    raise NotImplementedError("decode_nonce_authorize not implemented")


def sys_program_id() -> PublicKey:
    """Public key that identifies the System program."""
    return PublicKey("11111111111111111111111111111111")


def create_account(param: CreateAccountParams) -> Transaction:
    """Generate a Transaction that creates a new account."""
    raise NotImplementedError("create_account not implemented")


def assign(params: Union[AssignParams, AssignWithSeedParams]) -> Transaction:
    """Generate a Transaction that assigns an account to a program."""
    raise NotImplementedError("assign not implemented")


def transfer(params: TransferParams) -> Transaction:
    """Generate a Transaction that transfers lamports from one account to another.

    >>> from solana.publickey import PublicKey
    >>> sender, reciever = PublicKey(1), PublicKey(2)
    >>> transfer_tx = transfer(
    ...     TransferParams(from_pubkey=sender, to_pubkey=reciever, lamports=1000)
    ... )
    >>> type(transfer_tx)
    <class 'solana.transaction.Transaction'>
    """
    layout = SYSTEM_INSTRUCTION_LAYOUTS[_TRANSFER]
    data = encode_data(layout, params.lamports)

    txn = Transaction()
    txn.add(
        TransactionInstruction(
            keys=[
                AccountMeta(pubkey=params.from_pubkey, is_signer=True, is_writable=True),
                AccountMeta(pubkey=params.to_pubkey, is_signer=False, is_writable=True),
            ],
            program_id=sys_program_id(),
            data=data,
        )
    )
    return txn


def create_account_with_seed(params: CreateAccountWithSeedParams) -> Transaction:
    """Generate a Transaction that creates a new account at an address."""
    raise NotImplementedError("create_account_with_seed not implemented")


def create_nonce_account(param: Union[CreateNonceAccountParams, CreateAccountWithSeedParams]) -> Transaction:
    """Generate a Transaction that creates a new Nonce account."""
    raise NotImplementedError("create_nonce_account_params not implemented")


def nonce_initialization(param: InitializeNonceParams) -> TransactionInstruction:
    """Generate an instruction to initialize a Nonce account."""
    raise NotImplementedError("nonce_initialization not implemented")


def nonce_advance(param: AdvanceNonceParams) -> TransactionInstruction:
    """Generate an instruction to advance the nonce in a Nonce account."""
    raise NotImplementedError("nonce advance not implemented")


def nonce_withdraw(param: WithdrawNonceParams) -> Transaction:
    """Generate a Transaction that withdraws lamports from a Nonce account."""
    raise NotImplementedError("nonce_withdraw not implemented")


def nonce_authorize(param: AuthorizeNonceParams) -> Transaction:
    """Generate a Transaction that authorizes a new PublicKey as the authority on a Nonce account."""
    raise NotImplementedError("nonce_authorize not implemented")


def allocate(param: Union[AllocateParams, AllocateWithSeedParams]) -> Transaction:
    """Generate a Transaction that allocates space in an account without funding."""
    raise NotImplementedError("allocate not implemented")



"""Library to interface with Solana public keys."""
from __future__ import annotations

from typing import Any, List, Optional, Tuple, Union

import base58


class PublicKey:
    """The public key of a keypair.

    >>> # An arbitary public key:
    >>> pubkey = PublicKey(1)
    >>> str(pubkey) # String representation in base58 form.
    '11111111111111111111111111111112'
    >>> bytes(pubkey).hex()
    '0000000000000000000000000000000000000000000000000000000000000001'
    """

    LENGTH = 32
    """Constant for standard length of a public key."""

    def __init__(self, value: Union[bytearray, bytes, int, str, List[int]]) -> None:
        """Init PublicKey object."""
        self._key: Optional[bytes] = None
        if isinstance(value, str):
            try:
                self._key = base58.b58decode(value)
            except ValueError as err:
                raise ValueError("invalid public key input:", value) from err
            if len(self._key) != self.LENGTH:
                raise ValueError("invalid public key input:", value)
        elif isinstance(value, int):
            self._key = bytes([value])
        else:
            self._key = bytes(value)

        if len(self._key) > self.LENGTH:
            raise ValueError("invalid public key input:", value)

    def __bytes__(self) -> bytes:
        """Public key in bytes."""
        if not self._key:
            return bytes(self.LENGTH)
        return self._key if len(self._key) == self.LENGTH else self._key.rjust(self.LENGTH, b"\0")

    def __eq__(self, other: Any) -> bool:
        """Equality definition for PublicKeys."""
        return False if not isinstance(other, PublicKey) else bytes(self) == bytes(other)

    def __repr__(self) -> str:
        """Representation of a PublicKey."""
        return str(self)

    def __str__(self) -> str:
        """String definition for PublicKey."""
        return self.to_base58().decode("utf-8")

    def to_base58(self) -> bytes:
        """Public key in base58."""
        return base58.b58encode(bytes(self))

    def create_with_seed(self, from_public_key: PublicKey, seed: str, program_id: PublicKey) -> PublicKey:
        """Derive a public key from another key, a seed, and a program ID."""
        raise NotImplementedError("create_with_seed not implemented")

    def create_program_address(self, seeds: Union[bytearray, List[bytes]], program_id: PublicKey) -> PublicKey:
        """Derive a program address from seeds and a program ID."""
        raise NotImplementedError("create_program_address not implemented")

    def find_program_address(
        self, seeds: Union[bytearray, List[bytes]], program_id: PublicKey
    ) -> Tuple[PublicKey, int]:
        """Find a valid program address.

        Valid program addresses must fall off the ed25519 curve.  This function
        iterates a nonce until it finds one that when combined with the seeds
        results in a valid program address.
        """
        raise NotImplementedError("find_program_address not implemented")



"""Friendly JSON serializer & deserializer for Requests."""
import collections.abc
import json
from typing import Any, Dict, Iterable, Optional, Type


# Original source:
# https://github.com/ethereum/web3.py/blob/8bb0b56f65f1ab96b46db79f9956517ac628bd7b/web3/_utils/encoding.py#L181
class FriendlyJsonSerde:
    """Friendly JSON serializer & deserializer.

    When encoding or decoding fails, this class collects
    information on which fields failed, to show more
    helpful information in the raised error messages.
    """

    def _json_mapping_errors(self, mapping: Dict[Any, Any]) -> Iterable[str]:
        for key, val in mapping.items():
            try:
                self._friendly_json_encode(val)
            except TypeError as exc:
                yield "%r: because (%s)" % (key, exc)

    def _json_list_errors(self, iterable: Iterable[Any]) -> Iterable[str]:
        for index, element in enumerate(iterable):
            try:
                self._friendly_json_encode(element)
            except TypeError as exc:
                yield "%d: because (%s)" % (index, exc)

    @staticmethod
    def _is_list_like(obj: Any) -> bool:
        return not isinstance(obj, (bytes, str, bytearray)) and isinstance(obj, collections.abc.Sequence)

    def _friendly_json_encode(self, obj: Dict[Any, Any], cls: Optional[Type[json.JSONEncoder]] = None) -> str:
        try:
            encoded = json.dumps(obj, cls=cls)
            return encoded
        except TypeError as full_exception:
            if hasattr(obj, "items"):
                item_errors = "; ".join(self._json_mapping_errors(obj))
                raise TypeError("dict had unencodable value at keys: {{{}}}".format(item_errors)) from full_exception
            if FriendlyJsonSerde._is_list_like(obj):
                element_errors = "; ".join(self._json_list_errors(obj))
                raise TypeError("list had unencodable value at index: [{}]".format(element_errors)) from full_exception
            raise full_exception

    def json_decode(self, json_str: str) -> Dict[Any, Any]:  # pylint: disable=no-self-use
        """Deserialize JSON document to a Python object with friendly error messages."""
        try:
            decoded = json.loads(json_str)
            return decoded
        except json.decoder.JSONDecodeError as exc:
            err_msg = "Could not decode {} because of {}.".format(repr(json_str), exc)
            # Calling code may rely on catching JSONDecodeError to recognize bad json
            # so we have to re-raise the same type.
            raise json.decoder.JSONDecodeError(err_msg, exc.doc, exc.pos)

    def json_encode(self, obj: Dict[Any, Any], cls: Optional[Type[json.JSONEncoder]] = None) -> str:
        """Serialize obj to a JSON formatted `str` with friendly error messages."""
        try:
            return self._friendly_json_encode(obj, cls=cls)
        except TypeError as exc:
            raise TypeError("Could not encode to JSON") from exc
